"""Product block for :class:`Router` products."""

import ipaddress

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum
from pydantic import ConstrainedInt

from gso.products.product_blocks.site import (
    SiteBlock,
    SiteBlockInactive,
    SiteBlockProvisioning,
)


class RouterVendor(strEnum):
    """Enumerator for the different product vendors that are supported."""

    JUNIPER = "juniper"
    NOKIA = "nokia"


class RouterRole(strEnum):
    """Enumerator for the different types of routers."""

    P = "p"
    PE = "pe"
    AMT = "amt"


class PortNumber(ConstrainedInt):
    """Constrained integer for valid port numbers.

    The range from 49152 to 65535 is marked as ephemeral, and can therefore not be selected for permanent allocation.
    """

    gt = 0
    le = 49151


class RouterBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="RouterBlock",
):
    """A router that's being currently inactive. See :class:`RouterBlock`."""

    router_fqdn: str | None = None
    router_ts_port: PortNumber | None = None
    router_access_via_ts: bool | None = None
    router_lo_ipv4_address: ipaddress.IPv4Address | None = None
    router_lo_ipv6_address: ipaddress.IPv6Address | None = None
    router_lo_iso_address: str | None = None
    router_si_ipv4_network: ipaddress.IPv4Network | None = None
    router_ias_lt_ipv4_network: ipaddress.IPv4Network | None = None
    router_ias_lt_ipv6_network: ipaddress.IPv6Network | None = None
    router_vendor: RouterVendor | None = None
    router_role: RouterRole | None = None
    router_site: SiteBlockInactive | None
    router_is_ias_connected: bool | None = None


def generate_fqdn(hostname: str, site_name: str, country_code: str) -> str:
    """Generate an :term:`FQDN` from a hostname, site name, and a country code."""
    return f"{hostname}.{site_name.lower()}.{country_code.lower()}.geant.net"


class RouterBlockProvisioning(RouterBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A router that's being provisioned. See :class:`RouterBlock`."""

    router_fqdn: str
    router_ts_port: PortNumber
    router_access_via_ts: bool | None = None
    router_lo_ipv4_address: ipaddress.IPv4Address | None = None
    router_lo_ipv6_address: ipaddress.IPv6Address | None = None
    router_lo_iso_address: str | None = None
    router_si_ipv4_network: ipaddress.IPv4Network | None = None
    router_ias_lt_ipv4_network: ipaddress.IPv4Network | None = None
    router_ias_lt_ipv6_network: ipaddress.IPv6Network | None = None
    router_vendor: RouterVendor | None = None
    router_role: RouterRole | None = None
    router_site: SiteBlockProvisioning | None
    router_is_ias_connected: bool | None = None


class RouterBlock(RouterBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A router that's currently deployed in the network."""

    #:  :term:`FQDN` of a router.
    router_fqdn: str
    #:  The port of the terminal server that this router is connected to. Used to offer out of band access.
    router_ts_port: PortNumber
    #:  Whether this router should be accessed through the terminal server, or through its loopback address.
    router_access_via_ts: bool
    #:  The IPv4 loopback address of the router.
    router_lo_ipv4_address: ipaddress.IPv4Address
    #:  The IPv6 loopback address of the router.
    router_lo_ipv6_address: ipaddress.IPv6Address
    #:  The :term:`ISO` :term:`NET` of the router, used for :term:`ISIS` support.
    router_lo_iso_address: str
    #:  The SI IPv4 network of the router.
    router_si_ipv4_network: ipaddress.IPv4Network | None
    #:  The :term:`IAS` LT IPv4 network of the router.
    router_ias_lt_ipv4_network: ipaddress.IPv4Network | None
    #:  The :term:`IAS` LT IPv6 network of the router.
    router_ias_lt_ipv6_network: ipaddress.IPv6Network | None
    #:  The vendor of the router, can be any of the values defined in :class:`RouterVendor`.
    router_vendor: RouterVendor
    #:  The role of the router, which can be any of the values defined in :class:`RouterRole`.
    router_role: RouterRole
    #:  The :class:`Site` that this router resides in. Both physically and computationally.
    router_site: SiteBlock
    #:  The router is going to have an LT interface between inet0 and :term:`IAS`.
    router_is_ias_connected: bool
